"Python3 library to handle rectangular data grids (as from R)."

##  GRIDLIB :
##  Python3 data-grid routines :
##  Copyright 2012-2104, R.S. Forsyth.
##  Released under GNU Public licence, version 3:
##  http://www.gnu.org/licenses/gpl.txt
##  first version : 15/09/2012
##  last revision : 01/09/2014


ShowGNU = 0  ##  change to 1 2c licence statement
GNUblurb = """
  This copyrighted program is free software: you can redistribute it &/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.
"""
if ShowGNU:
    print(GNUblurb)
    
import  math, os, string, sys
##  import  time, random


if sys.platform.startswith("linu"):  ##  presume linux
    sys.path.append("~/py3libs")
else:  ##  presume windows
    sys.path.append("c:\\py3libs")
##  in case executed from somewhere else.


##  local libraries :
from globdefs import *
from glib import  numeric, numerize


##  COLS2USE :
def cols2use (paradat,numsonly=True):
    "works out which columns to use in calculations."

    goodcols = set(range(paradat.cols))
    ##  lowtags = [t.lower() for t in paradat.taglist]
    ##  first remove non-numeric columns :
    for vn,varname in enumerate(paradat.taglist):
        nn = paradat.tagdict[varname]
        if numsonly and (nn < paradat.rows):
            goodcols.discard(vn)
    ##  don't remove target column at this stage :
    ##  goodcols.discard(paradat.targcol)
    ##  also remove vars that user asks to omit :
    if paradat.skipvars:
        bits = paradat.skipvars.split(',')  ##  comma-delimited
        for b in bits:
            bb = b.strip()
            if bb in paradat.taglist:
                bcol = paradat.taglist.index(bb)
                goodcols.discard(bcol)
            else:
                print("Warning : skipvar",b,"not found in taglist!")

    return  goodcols  ##  set of usable numeric cols.


##  MATDUMP :
def matdump (fo,dmat):
    "dumps numeric square matrix onto o/p file."

    n = len(dmat)
    lo = 0
    for dvec in dmat:
        lout = vecline(dvec)
        fo.write(lout + NL)
        lo += 1
    assert  lo == n
    
    return  n


##  VECLINE :
def vecline (datavec,decimals=5,dlim=Htab):
    "returns a delimited string from a list of numerics."

    lout = ''
    for v in datavec:
        if type(v) == type(9):
            sv = str(v)
        elif type(v) == type(9.9):
            sv = str(round(v,decimals))
        else:  ##  something wrong ?
            sv = str(v)
        lout = lout + sv + dlim

    return  lout[:-1]  ##  remove trailing delimiter


##  MATREAD :
def matread (filename,paradat,tabchar=Htab):
    "reads data grid with header as first line."

    fi = open(filename,'r',encoding="utf-8")
    ##  crash if failure!
    datlines = fi.readlines()
    h = datlines[0].strip()  ##  presumes header line first
    lc = len(datlines)
    taglist = h.split(tabchar)  ##  assumes tab as delimiter
    cols = len(taglist)
    print(lc,cols)
    ##  print(taglist)
    ##  make dictionary 4 guessing column types :
    tagdict = {}
    for t in taglist:
        tagdict[t] = 0

    datamat = []
    for r in range(1,lc):
        datlist = datlines[r].strip().split(tabchar)
        vals = len(datlist)
        if vals < 1:
            continue  ##  skip blank lines
        if vals < cols:
            print("Warning: dodgy data line :")
            print("items =",vals,"vals = ",datlist)
            return  datamat, taglist  ##  will force early exit
        rowdata = []
        for tp,item in enumerate(datlist):
            v = numerize(item)  ##  most values expected 2b numeric
            if numeric(v):
                rowdata.append(v)
                tagdict[taglist[tp]] += 1
            else:
                rowdata.append(item)  ##  probably string value
                tagdict[taglist[tp]] -= 1
        datamat.append(rowdata)
        
    return  (datamat,taglist,tagdict)


##  MAKEFULL :
def makefull (prepath,filename):
    "joins prepath to filename, with separator if missing."

    if prepath.endswith(os.sep):
        fullname = prepath + filename
    else:  ##  insert separator
        fullname = prepath + os.sep + filename

    return  fullname


##  VAR2COL :
def var2col (varname,varnames,goodcols=set()):
    "tries to find varname in varnames, only in goodcols if given."

    if not goodcols:
        goodcols = set(range(len(varnames)))
    if varname in varnames:
        vpos = varnames.index(varname)
        if vpos in goodcols:
            return  vpos  ##  position of varname

    return  -1  ##  failure signal


##  BASTATZ :
def bastatz (numlist):
    "basic stats routine for a list of numbers."

    fn = float(len(numlist))
    sum1 = 0.0
    sum2 = 0.0
    for x in numlist:
        sum1 += x
        sum2 += x**2

    statsack = Sack(0)
    ##  defaults for dodgy inputs :
    statsack.mean = sum1
    statsack.var1 = statsack.var2 = 0.0
    statsack.std1 = statsack.std2 = 0.0
    if fn > 0.0:
        statsack.mean = sum1 / fn
        statsack.var1 = sum2 / fn - statsack.mean**2
        ##  print "$$",statsack.var1
        statsack.std1 = math.sqrt(max(0.0,statsack.var1))
        if fn > 1.0:
            statsack.var2 = statsack.var1 * (fn/(fn-1.0))
            statsack.std2 = math.sqrt(max(0.0,statsack.var2))

    return statsack


##  MIDVALUE :
def midvalue (xvec):
    "simple median."

    vecsize = len(xvec)
    if vecsize < 1:
        return  0.0  ##  None better ?
    if vecsize % 2 == 1:  ##  uneven
        m = vecsize // 2
        return  sorted(xvec)[m]
    else:  ##  even
        m = (vecsize-1)//2
        v1,v2 = sorted(xvec)[m:m+2]
        return  (v1+v2)*0.5
